var {
  THREE,
  Edge,
  Node,
  OrbitControls,
  d3,
  ee,
  DEF_NODE_COLOR,
  DEF_NODE_HIGHLIGHT,
  DEF_ADJACENT_HIGHLIGHT,
  DEF_NODE_HTEXTURE,
  DEF_NODE_TEXTURE,
  DEF_NODE_NO_HTEX,
  DEF_NODE_NO_TEX,
  DEF_NODE_NO_HCOLOR,
  DEF_NODE_NO_COLOR,
  DEF_NODE_NO_HIGHLIGHT,
  DEF_NODE_SHAPE,
  DEF_NODE_HSHAPE,
  DEF_NODE_SIZE,
  DEF_LAYOUT,
  DEF_LINES,
  DEF_MAP,
  DEF_TEXT_SCALE,
  DEF_TEXT_SIZE,
  DEF_TEXT_X_OFFSET,
  DEF_TEXT_Y_OFFSET,
  DEF_TEXT_BACK_COLOR,
  DEF_TEXT_OPACITY,
  TEXT_WIREFRAME
} = require('./imports').default;
var Utils = require('./utils/utils');

module.exports = function (self) {
  /**
   *  Create highlight mesh
   */
  self.createHighlightMesh = function (node, texture) {
    var nsize;
    if (node.renderData.size != undefined) nsize = node.renderData.size;
    else nsize = node.renderData['size'];

    var shape = node.renderData.hshape || node.renderData.shape || DEF_NODE_HSHAPE;
    var size = nsize * 1.5;
    if (shape == "square") {
      var rectShape = new THREE.Shape();
      rectShape.moveTo(-1, -1);
      rectShape.lineTo(-1, 1);
      rectShape.lineTo(1, 1);
      rectShape.lineTo(1, -1);
      rectShape.lineTo(-1, -1);
      var hgeometry = new THREE.ShapeGeometry(rectShape);
    } else {
      var hgeometry = new THREE.CircleGeometry(1, 32);
    }
    texture = DEF_NODE_NO_HTEX ? null : texture;
    var color = DEF_NODE_NO_HCOLOR ? null : parseInt(node.renderData.hcolor || DEF_NODE_HIGHLIGHT);
    var hmaterial = new THREE.MeshBasicMaterial({color: color, map: texture});
    hmaterial.transparent = true;
    var highlight = new THREE.Mesh(hgeometry, hmaterial);
    highlight.visible = false;
    highlight.scale.set(1.25, 1.25, 1);
    highlight.position.z = -0.01;
    return highlight;
  }

  /**
   *  Create node mesh
   */
  self.createNodeMesh = function (node, texture) {
    var size;
    if (node.renderData.size != undefined) size = node.renderData.size;
    else size = node.renderData['size'];

    var shape = node.renderData.shape || DEF_NODE_SHAPE;
    if (shape == "square") {
      var rectShape = new THREE.Shape();
      rectShape.moveTo(-1, -1);
      rectShape.lineTo(-1, 1);
      rectShape.lineTo(1, 1);
      rectShape.lineTo(1, -1);
      rectShape.lineTo(-1, -1);
      var geometry = new THREE.ShapeGeometry(rectShape);
    } else {
      var geometry = new THREE.CircleGeometry(1, 32);
    }
    texture = DEF_NODE_NO_TEX ? null : texture;
    var color = DEF_NODE_NO_COLOR ? null : new THREE.Color(node.renderData.color || DEF_NODE_COLOR);
    var material = new THREE.MeshBasicMaterial({color: color, map: texture})
    material.transparent = true;
    var draw_object = new THREE.Mesh(geometry, material);
    draw_object.scale.set(size, size, 1);
    return draw_object;
  }

  /**
   *  Create label sprite
   */
  self.createLabelSprite = function (node) {
    // Create canvas to draw text label onto sprite image
    var canvas = document.createElement('canvas');
    var context = canvas.getContext('2d');
    context.font = node.renderData.font || "Bold 32px Arial";
    var metrics = context.measureText(node.renderData.label || "");
    var textWidth = metrics.width;
    context.fillStyle = "rgba(0, 0, 0, 1.0)";
    context.fillText(node.renderData.label || "", 0, 32);
    var texture1 = new THREE.Texture(canvas);
    texture1.needsUpdate = true;
    var material1 = new THREE.SpriteMaterial({map: texture1});
    var sprite = new THREE.Sprite(material1);
    // Scale and adjust label
    sprite.scale.set(10, 5, 1.0);
    sprite.position.x += 6.5;
    sprite.position.y -= 1.75;
    return sprite;
  }

  /**
   *  Create label sprite
   */
  self.createLabelDiv = function (node) {
    // Create canvas to draw text label onto sprite image
    var element = document.createElement("div");
    element.style.width = '1000px';
    element.style.height = '200px';
    element.style.backgroundColor = DEF_TEXT_BACK_COLOR;
    element.style.opacity = DEF_TEXT_OPACITY;
    var content = document.createTextNode(node.renderData.label);
    var textElement = document.createElement("div");
    textElement.style.width = '1000px';
    textElement.style.height = '200px';
    textElement.style.fontSize = "96px";
    textElement.style.textOverflow = "ellipsis";
    textElement.style.overflow = "hidden";
    textElement.appendChild(content);
    element.appendChild(textElement);
    var labelObj = new THREE.CSS3DObject( element );
    return labelObj;
  }

  /**
   *  Create a node object and add it to the scene.
   */
  self.drawNode = function (node) {
    function onHTextureLoad(node, htexture) {
      // Load node texture
      self.textureLoader.load(node.renderData.image || DEF_NODE_TEXTURE,
        onTextureLoad.bind(null, node, htexture));
    }

    function onTextureLoad(node, htexture, texture) {
      // Create mesh for node
      var draw_object = self.createNodeMesh(node, texture);

      // Create highlight
      if (!DEF_NODE_NO_HIGHLIGHT) {
        var highlight = self.createHighlightMesh(node, htexture);
        draw_object.add(highlight);
      }

      // Create div for label
      var label = self.createLabelDiv(node);
      var textHolder = new THREE.Object3D();
      textHolder.add(label);
			var material = new THREE.MeshBasicMaterial( { color: 0x000000, wireframe: true, wireframeLinewidth: 1, side: THREE.DoubleSide } );
      var geometry = new THREE.PlaneGeometry(1000, 200);
      var textMesh = new THREE.Mesh(geometry, material);
      textMesh.scale.set(DEF_TEXT_SIZE, DEF_TEXT_SIZE, 1);
      var bbox = new THREE.Box3().setFromObject(textMesh);
      textMesh.position.x += (bbox.max.x - bbox.min.x) * DEF_TEXT_X_OFFSET;
      textMesh.position.y -= (bbox.max.y - bbox.min.y) * DEF_TEXT_Y_OFFSET;
      label.element.hidden = true;
      textMesh.visible = TEXT_WIREFRAME;
      label.position.copy(textMesh.position);
      label.scale.copy(textMesh.scale);
      textHolder.add(textMesh);
      node.renderData.textHolder = textHolder;
      self.scene.add(textHolder);

      // Set location of node to given locaton
      draw_object.position.x = node.renderData.x;
      draw_object.position.y = node.renderData.y;
      draw_object.index = node.id;
      node.renderData.draw_object = draw_object;

      self.nodeCount += 1;
      self.nodes.add(draw_object);
    }

    if (DEF_NODE_NO_HTEX && DEF_NODE_NO_TEX) {
      onTextureLoad(node, null, null);
    } else {
      // Load highlight texture
      self.textureLoader.load(node.renderData.himage || DEF_NODE_HTEXTURE,
        onHTextureLoad.bind(null, node));
    }

  };

  /**
   *  Create an edge and add it to the lineSegments.
   */
  self.drawEdge = function (source, target) {
    if (source.renderData.draw_object === undefined || target.renderData.draw_object === undefined) {
      setTimeout(self.drawEdge.bind(null, source, target), 50);
      return;
    }
    if (DEF_LINES == "fancy") {
      var geometry = new THREE.Geometry();
      geometry.vertices.push(new THREE.Vector2(source.x, source.y));
      geometry.vertices.push(new THREE.Vector2(target.x, target.y));
      var line = new THREE.MeshLine();
      line.setGeometry(geometry);
      var material = new THREE.MeshLineMaterial({
        useMap: false,
        color: source.renderData.linecolor,
        opacity: 1,
        resolution: self.resolution,
        sizeAttenuation: true,
        lineWidth: 1,
        near: self.ccamera.near,
        far: self.ccamera.far
      });
      var mesh = new THREE.Mesh(line.geometry, material);
      mesh.frustumCulled = false;
      self.edges.push(line);
      self.scene.add(mesh);
      self.lineObjects[(self.edges.length * 2) - 2] = source;
      self.lineObjects[(self.edges.length * 2) - 1] = target;
      self.lineIndices.push({"source": source, "target": target, "hide": false});
    } else {
      var positions = self.edges.attributes.position.array;
      var colors = self.edges.attributes.color.array;
      var index = self.drawCount * 3;
      self.lineObjects[(index / 3) - 2] = source;
      self.lineObjects[(index / 3) - 1] = target;
      if (!source.linkObjs) {
        source.linkObjs = [];
      }
      if (!target.linkObjs) {
        target.linkObjs = [];
      }
      self.lineIndices.push({"source": source, "target": target, "hide": false, "linecolor": source.renderData.linecolor});
      source.linkObjs.push(self.lineIndices[(index / 6)]);
      target.linkObjs.push(self.lineIndices[(index / 6)]);
      var v1pos = source.renderData.draw_object.position;
      var v2pos = target.renderData.draw_object.position;
      var v1color = self.lineIndices[(index / 6)].linecolor;
      var v2color = self.lineIndices[(index / 6)].linecolor;
      colors[index] = v1color.b;
      positions[index++] = v1pos.x;
      colors[index] = v1color.g;
      positions[index++] = v1pos.y;
      colors[index] = v1color.r;
      positions[index++] = v1pos.z - 0.02;
      colors[index] = v2color.b;
      positions[index++] = v2pos.x;
      colors[index] = v2color.g;
      positions[index++] = v2pos.y;
      colors[index] = v2color.r;
      positions[index++] = v2pos.z - 0.02;
      self.drawCount += 2;
    }
  };
}